// *****************************************************************************
//
// Copyright. 2007. Symvionics, Inc. All rights reserved.
// 
// Symvionics Inc. ("Symvionics") grants a non-exclusive license for you to use
// and redistribute this software, in both source code and binary forms, and the
// accompanying documentation, if any, free of charge, provided that this
// paragraph, and the following two paragraphs appear in all copies and in all
// derivative works.  In no event shall this source code or this license thereto
// impact upon or fall within the scope of any otheragreement that you have with
// Symvionics.
// 
// BY USING THIS SOFTWARE, YOU FURTHER AGREE THAT, IN NO EVENT, SHALL SYMVIONICS
// BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
// SOFTWARE, THE ACCOMPANYING DOCUMENTATION, IF ANY, OR ANY DERIVATIVE WORKS.
// YOU AGREE TO DEFEND, INDEMNIFY, AND HOLD HARMLESS SYMVIONICS FROM AND AGAINST
// ALL FINES, PENALTIES, LIABILITIES, DAMAGES, COSTS, AND EXPENSES INCURRED
// BY SYMVIONICS AS A RESULT OF YOUR USE OF THIS SOFTWARE INCLUDING THIRD PARTY
// USE OF ANY DERIVATIVE WORKS CREATED BY YOU.
// 
// SYMVIONICS SPECIFICALLY DISCLAIMS ANY AND ALL WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE.  THE SOFTWARE AND ACCOMPANYING DOCUMENTATION LICENSED
// HEREUNDER ARE PROVIDED "AS IS" WITHOUT WARRANTY OR OBLIGATION OF SUPPORT OR 
// MAINTENANCE.
//
// *****************************************************************************

#ifndef _OGLHelperTemplates_
#define _OGLHelperTemplates_

#include "gl/gl.h"
#include "gl/glaux.h" // Could use this if needed..
#include <limits.h>  // For MaxInt

#pragma comment (lib, "opengl32.lib")  // OpenGL libraries to link
#pragma comment (lib, "glu32.lib")


// Some OGL helper templates...
template <class DWORDType> bool tIsSofwareGL( DWORDType d )
{ 
  bool genacc = (d&PFD_GENERIC_ACCELERATED)!=0;
  bool genfmt = (d&PFD_GENERIC_FORMAT)!=0;
  if( genfmt && !genacc ) return true;
  return false;
}

template <class PIXELFORMATDESCRIPTORType> void tOutputPFDCapabilities( PIXELFORMATDESCRIPTORType* pfd, char* destString = NULL )
{
  if ( !pfd ) return;

  char s[250];

  DWORD d = pfd->dwFlags;

  bool genacc = (d&PFD_GENERIC_ACCELERATED)!=0;
  bool genfmt = (d&PFD_GENERIC_FORMAT)!=0;
  if (!genfmt && !genacc) strcpy( s, "Hardware Accel (ICD) in GL - " );
  if (!genfmt &&  genacc) strcpy( s, "Unknown GL Accel - " );
  if ( genfmt && !genacc) strcpy( s, "Software GL - " );
  if ( genfmt &&  genacc) strcpy( s, "Hardware Accel (MCD) in GL - " );

  if( d&PFD_DRAW_TO_WINDOW )      strcpy( s,"draw-to-window,") ;
  if( d&PFD_DRAW_TO_BITMAP )      strcpy( s,"draw-to-bitmap,") ;
  if( d&PFD_SUPPORT_GDI )         strcpy( s,"support-gdi,") ;
  if( d&PFD_SUPPORT_OPENGL )      strcpy( s,"support-opengl,") ;
  if( d&PFD_GENERIC_ACCELERATED ) strcpy( s,"generic-acc,") ;
  if( d&PFD_GENERIC_FORMAT )      strcpy( s,"generic-fmt,") ;
  if( d&PFD_NEED_PALETTE )        strcpy( s,"need-pal,") ;
  if( d&PFD_NEED_SYSTEM_PALETTE ) strcpy( s,"need-sys-pal,") ;
  if( d&PFD_DOUBLEBUFFER )        strcpy( s,"dblbuff,") ;
  if( d&PFD_STEREO )              strcpy( s,"stereo,") ;
  if( d&PFD_SWAP_LAYER_BUFFERS )  strcpy( s,"swap-layer-bufs,") ;

  char buf[500];
  sprintf( buf, " - CB:%d DB:%d\n", pfd->cColorBits, pfd->cDepthBits ); 
  strcat( s, buf );

  if ( destString )
    strcpy( destString, s );
  else
    ::OutputDebugString( s );
}

template <class HDCType> HRESULT tSetupOGLPixelFormat( HDCType hdc, BOOL hardwareAccel, BOOL doubleBuffered, BOOL verticalSync, int depthBits = 16, int colorBits = 16, PIXELFORMATDESCRIPTOR* returnPfd = NULL )
{
  // These define our target pixel format
  int OGLPixelType = PFD_TYPE_RGBA;
  int OGLLayerType = PFD_MAIN_PLANE; 
  PIXELFORMATDESCRIPTOR pfd = { sizeof(PIXELFORMATDESCRIPTOR), 1,
                                (PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER), (BYTE)PFD_TYPE_RGBA,
                                (BYTE)colorBits, (BYTE)0, (BYTE)0, (BYTE)0, (BYTE)0, (BYTE)0, (BYTE)0, (BYTE)0, (BYTE)0, (BYTE)0,
                                (BYTE)0, (BYTE)0, (BYTE)0, (BYTE)0, (BYTE)depthBits, (BYTE)0, (BYTE)0, PFD_MAIN_PLANE, 0, 0, 0, 0 };

  // If the user asked for hardware acceleration then go through the process of
  // getting the closest match to our target pixel format
  // Debug: Load target pixel format from outside the program ??
  if( hardwareAccel )
  {
    // Override target with user selectable options from the command line
    pfd.cColorBits = (BYTE)colorBits;  
    pfd.cDepthBits = (BYTE)depthBits;  
    if( doubleBuffered )
      pfd.dwFlags |= PFD_DOUBLEBUFFER;

    // Find the index of the closest pixel format to our target format
    int PixelFormat = ChoosePixelFormat( hdc, &pfd );
    
    if( PixelFormat == 0 ) PixelFormat = 1; // What to do ?

    // Get the actual pfd settings for this closest match - set the global guy so all will get it
    if( DescribePixelFormat( hdc, PixelFormat, sizeof(pfd), &pfd ) > 0 )
    {
      if( SetPixelFormat(hdc, PixelFormat, &pfd) == TRUE )
      {
        if ( returnPfd ) memcpy( returnPfd, &pfd, sizeof(PIXELFORMATDESCRIPTOR) );
        OutputDebugString( "User Request Hardware Acceleration...\n" );
        tOutputPFDCapabilities( &pfd );
        return S_OK;
      }
    }
  }

  // Attempt software mode
  ZeroMemory( &pfd, sizeof(pfd) ) ;
  int nFormats = DescribePixelFormat( hdc, 1, sizeof(pfd), &pfd );

  // Try to get a match where the color and depth bits are the lowest possible
  // size if the user specifies the depth or color bits to be zero
  if( colorBits == 0 || depthBits == 0 )
  {
    int score = INT_MAX;
    int PixelFormat = 0;

    for( int n=1; n <= nFormats;  n++ )
    {
      if( DescribePixelFormat( hdc, n, sizeof(pfd), &pfd ) == 0 ) continue; // If Fail - try the next format

      // Find the minimal depth and color bits if either of these are zero
      if( tIsSofwareGL(pfd.dwFlags)            && 
            pfd.iPixelType == OGLPixelType   &&
            pfd.iLayerType == OGLLayerType   &&
            (pfd.dwFlags&PFD_SUPPORT_OPENGL)  &&
            (pfd.dwFlags&PFD_DRAW_TO_WINDOW)  &&
            (pfd.dwFlags&PFD_DOUBLEBUFFER) )
      {
        // Debug: not checking if color bits are too little - but it didn't seem to matter ??
        if( (pfd.cColorBits+pfd.cDepthBits) < score )
        {
          score = pfd.cColorBits+pfd.cDepthBits ;
          PixelFormat = n ;
        }
      }
    }
    if( SetPixelFormat(hdc, PixelFormat, &pfd) == TRUE )
    {
      if ( returnPfd ) memcpy( returnPfd, &pfd, sizeof(PIXELFORMATDESCRIPTOR) );
      OutputDebugString( "Software - Minimize color/depth bits...\n" );
      tOutputPFDCapabilities( &pfd );
      return S_OK;
    }
  }

  // user asked for specfic depth/color settings attempt an exact match
  int PixelFormat = 0;
  int n;  // VS2005 'for' loop scope fix
  for( n=1; n <= nFormats;  n++)
  {
    if( DescribePixelFormat( hdc, n, sizeof(pfd), &pfd ) == 0 ) continue;

    if( tIsSofwareGL(pfd.dwFlags) && 
          pfd.cColorBits == colorBits &&
          pfd.cDepthBits == depthBits &&
          pfd.iPixelType == OGLPixelType   &&
          pfd.iLayerType == OGLLayerType   &&
          (pfd.dwFlags&PFD_SUPPORT_OPENGL)  &&
          (pfd.dwFlags&PFD_DRAW_TO_WINDOW)  &&
          (pfd.dwFlags&PFD_DOUBLEBUFFER) )
    {
      PixelFormat = n;
      if( SetPixelFormat(hdc, PixelFormat, &pfd) == TRUE )
      {
        if ( returnPfd ) memcpy( returnPfd, &pfd, sizeof(PIXELFORMATDESCRIPTOR) );
        OutputDebugString( "Software - Attempt exact match on color/depth bits\n" );
        tOutputPFDCapabilities( &pfd );
        return S_OK;
      }
    }
  }

  // We weren't able to do an exact match so try a looser (>=) standard.
  for( n=1; n <= nFormats;  n++)
  {
    if( DescribePixelFormat( hdc, n, sizeof(pfd), &pfd ) == 0 ) continue;

    // DEBUG: Check entire target PFD later - also get it from command line
    if( tIsSofwareGL(pfd.dwFlags) && 
          pfd.cColorBits >= colorBits &&
          pfd.cDepthBits >= depthBits &&
          pfd.iPixelType == OGLPixelType   &&
          pfd.iLayerType == OGLLayerType   &&
          (pfd.dwFlags&PFD_SUPPORT_OPENGL)  &&
          (pfd.dwFlags&PFD_DRAW_TO_WINDOW)  &&
          (pfd.dwFlags&PFD_DOUBLEBUFFER) )
    {
      PixelFormat = n ;
      if( SetPixelFormat(hdc, PixelFormat, &pfd) == TRUE )
      {
        if ( returnPfd ) memcpy( returnPfd, &pfd, sizeof(PIXELFORMATDESCRIPTOR) );
        OutputDebugString( "Software - Attempt any match >= to color/depth bits\n" );
        tOutputPFDCapabilities( &pfd );
        return S_OK;
      }
    }
  }

  // Well if all else fails try loading our target as if hardware mode was requested
  PixelFormat = ChoosePixelFormat( hdc, &pfd );
  if( PixelFormat == 0 ) PixelFormat = 1; // What to do ?

  // Get the actual pfd settings for this closest match - set the global guy so all will get it
  if( DescribePixelFormat( hdc, PixelFormat, sizeof(pfd), &pfd ) > 0 )
  {
    if( SetPixelFormat(hdc, PixelFormat, &pfd) == TRUE )
    {
      if ( returnPfd ) memcpy( returnPfd, &pfd, sizeof(PIXELFORMATDESCRIPTOR) );
      OutputDebugString( "User Request Hardware Acceleration...\n" );
      tOutputPFDCapabilities( &pfd );
      return S_OK;
    }
  }

  // Wow we really did fail after all those attempts
  return E_FAIL;
}

// Template only works with COLORREF or OLE_COLOR types for now..
template <class COLORREFType> void tSetColor( COLORREFType color )
{
  double red, green, blue;
  red   = GetRValue(color)/255.0;
  green = GetGValue(color)/255.0;
  blue  = GetBValue(color)/255.0;
  
  glColor3d( red>1.0?1.0:red, green>1.0?1.0:green, blue>1.0?1.0:blue );
}

#define PT     1
#define STROKE 2
#define END    3
#define BASE_NUMBER 128
typedef struct charpoint 
{
  GLfloat x, y;
  int     type;
} CP;

// Template only works with "char" type for now... (no wide characters)
template <class StringType> void tPrintStrokedString( StringType *s )
{
  GLsizei len = (int)strlen(s);
  glCallLists( len, GL_BYTE, (GLbyte *)s );
}

template <class CPType> void tDrawLetter( CPType* l )
{
  glEnable(GL_LINE_SMOOTH);
  glEnable(GL_BLEND);
  
  glBegin( GL_LINE_STRIP );
  while(1)
  {
    switch(l->type)
    {
    case PT:
      glVertex2fv(&l->x);
      break;
    case STROKE:
      glVertex2fv(&l->x);
      glEnd();
      glBegin( GL_LINE_STRIP );
      break;
    case END:
      glVertex2fv(&l->x);
      glEnd();
      glTranslatef(8.0, 0.0, 0.0);
      return;
    }
    l++;
  }
  
  glDisable(GL_LINE_SMOOTH);
  glDisable(GL_BLEND);
}

// Trim example to generate call lists for rendering numbers only...Call tPrintStrokedString
// I suggest using other code from the Net to render text.. 
// This is a poor example that uses assumes no other call lists are present..
static void tGenerateNumberCallLists( void )
{
  glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

  GLuint NumberBase;

  glColor3d( 0.0, 0.0, 0.0 );
  NumberBase = glGenLists( BASE_NUMBER );
  glListBase( NumberBase );  
  
  CP ZeroData[] = {
    {0, 2, PT}, {0, 7, PT}, {2, 9, PT}, {4, 9, PT}, {6, 7, PT}, {6, 2, PT},
    {4, 0, PT}, {2, 0, PT}, {0, 2, END} 
  };
  glNewList(NumberBase+'0', GL_COMPILE); tDrawLetter(ZeroData);
  glEndList();
  
  CP OneData[] = {
    {1, 0, PT}, {5, 0, STROKE}, {3, 0, PT}, {3, 9, PT}, {1, 7, END}
  };
  glNewList(NumberBase+'1', GL_COMPILE); tDrawLetter(OneData);
  glEndList();
  
  CP TwoData[] = {
    {0, 7, PT}, {2, 9, PT}, {4, 9, PT}, {6, 7, PT}, {6, 6, PT}, {2,3,PT},{0, 0, PT},
    {7, 0, END}
  };
  glNewList(NumberBase+'2', GL_COMPILE); tDrawLetter(TwoData);
  glEndList();
  
  CP ThreeData[] = {
    {0, 8, PT}, {1, 9, PT}, {5, 9, PT}, {6, 8, PT}, {6, 7, PT}, {4, 5, PT},
    {3, 5, PT}, {4, 5, PT}, {6, 3, PT}, {6, 1, PT}, {5, 0, PT}, {1, 0, PT},
    {0, 1, END}
  };
  glNewList(NumberBase+'3', GL_COMPILE); tDrawLetter(ThreeData);
  glEndList();
  
  CP FourData[] = {
    {4, 0, PT}, {4, 9, PT}, {0, 4, PT}, {6, 4, END}
  };
  glNewList(NumberBase+'4', GL_COMPILE); tDrawLetter(FourData);
  glEndList();
  
  CP FiveData[] = {
    {0, 1, PT}, {1, 0, PT}, {4, 0, PT}, {6, 2, PT}, {6, 3, PT}, {4, 5, PT},
    {1, 5, PT}, {1, 9, PT}, {6, 9, END}
  };
  glNewList(NumberBase+'5', GL_COMPILE); tDrawLetter(FiveData);
  glEndList();
  
  CP SixData[] = {
    {3, 9, PT}, {0, 3, PT}, {0, 2, PT}, {2, 0, PT}, {4, 0, PT}, {6, 2, PT},
    {6, 3, PT}, {4, 5, PT}, {2, 5, PT}, {0, 3, END}
  };
  glNewList(NumberBase+'6', GL_COMPILE); tDrawLetter(SixData);
  glEndList();
  
  CP SevenData[] = {
    {0, 9, PT}, {6, 9, PT}, {3, 5, PT}, {3, 0, END}
  };
  glNewList(NumberBase+'7', GL_COMPILE); tDrawLetter(SevenData);
  glEndList();
  
  CP EightData[] = {
    {0, 1, PT}, {0, 4, PT}, {1, 5, PT}, {5, 5, PT}, {6, 6, PT}, {6, 8, PT},
    {5, 9, PT}, {1, 9, PT}, {0, 8, PT}, {0, 6, PT}, {1, 5, PT}, {5, 5, PT},
    {6, 4, PT}, {6, 1, PT}, {5, 0, PT}, {1, 0, PT}, {0, 1, END}
  };
  glNewList(NumberBase+'8', GL_COMPILE); tDrawLetter(EightData);
  glEndList();
  
  CP NineData[] = {
    {6, 0, PT}, {6, 7, PT}, {4, 9, PT}, {2, 9, PT}, {0, 7, PT}, {0, 6, PT},
    {2, 4, PT}, {4, 4, PT}, {6, 5, END}
  };
  glNewList(NumberBase+'9', GL_COMPILE); tDrawLetter(NineData);
  glEndList();
  
  CP MinusData[] = { {1, 5, PT}, {5, 5, END} };
  glNewList(NumberBase+'-', GL_COMPILE); tDrawLetter(MinusData);
  glEndList();
  
 
}


#endif // _OGLHelperTemplates_

